/**
 * @mainpage GNUnet documentation
 *
 * @section Introduction
 *
 * See the [Modules](modules.html) section for documentation of the various services, APIs and libraries of GNUnet.
 *
 * Some additional documentation can be found under [Files](files.html), in the src/include/ directory.
 *
 * See also the [handbooks](https://docs.gnunet.org/) on installation, and for user and developer documentation.
 */
